{% extends "main.html" %}
{% block tabs %}
  {{ super() }}
  <section class="mdx-container">
    <div class="md-grid md-typeset">
      <div class="mdx-hero">
        <div class="mdx-hero__content">
          <h1>Welcome to the {{ config.site_name }}</h1>
          <p>{{ config.site_description }}</p>
          <a href="./{{ 'get-started/quick-start-guide' | url }}" title="Get Started" class="md-button md-button--primary md-button--teriary">
            Get started
          </a>
          <a href="./{{ 'get-started/get-your-planet-account' | url }}" title="Get your Planet account" class="md-button md-button--secondary">
            Get a Planet account
          </a>
        </div>
      </div>
    </div>
  </section>
{% endblock %}
{% block content %}{% endblock %}
<!-- More minimal footer -->
{% block footer %}
<footer class="md-footer">
  <div class="md-footer-meta__inner md-grid">
      <!-- Copyright and theme information -->
      <div class="md-footer-copyright">
          &copy; {{ build_date_utc.strftime('%Y') }} <a href='https://planet.com' title="Planet Labs PBC" target='_blank'>Planet Labs PBC</a>
      </div>
      <div class="md-footer-right">
          powered by
          <a href="https://www.mkdocs.org" title="MkDocs">MkDocs</a>
          and 
          <a href="https://squidfunk.github.io/mkdocs-material/"
            title="Material for MkDocs">
              Material for MkDocs</a>
      </div>    
  </div>
</footer>
{% endblock %}
